# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import logging
from functools import reduce
from .converters import CONVERTERS, convert_from_fiql


JSON_QUERY_COMPATIBLE = True
JSON_QUERY_SUBSTITUTIONS = {
    r'/%3C=/g': '=le=',
    r'/%3E=/g': '=ge=',
    r'/%3C/g': '=lt=',
    r'/%3E/g': '=gt='
}


PARSING_REGEX = r'(\))|([&\|,])?([\+\*\$\-:\w%\._]*)(\(?)'
BUBBLE_TO_TOP_OPERATORS = ['sort', 'select', 'values', 'limit']
PRIMARY_KEY_NUMBER = 'id'
DEFAULT_NODE_NAME = 'and'


class URIError(Exception):
    pass


class ParserError(Exception):
    pass


def str_to_val(string):
    converter = CONVERTERS['auto']

    if ':' in string:
        convert_name, rest = string.split(":", 1)
        try:
            logging.info('attempting to convert {} with {} converter'.format(string, convert_name))
            converter = CONVERTERS[convert_name]
        except KeyError as e:
            raise ParserError('Unknown converter ' + convert_name, e)
        string = ':'.join([rest])

    return converter(string)


class Query(object):
    def __init__(self, name='', args=None, parent=None):
        self.name = name
        self.args = args
        self.cache = {}
        self.parent = parent

        if self.args is None:
            self.args = []

    def add_argument(self, arg):
        self.args.append(arg)
        return self

    def set_conjuction(self, operator):
        if not self.name:
            self.name = operator
        elif self.name != operator:
            raise ParserError(
                'Can not mix conjunctions within a group, use paranthesis around each set of same conjuctions (& and |)'
            )
        return self

    def dict(self):
        return {'name': self.name, 'args': [x.dict() if isinstance(x, Query) else x for x in self.args]}

    def __repr__(self):
        return str(self.dict())


def parse(query):

    top_term = term = Query()

    if isinstance(query, Query):
        return query
    elif isinstance(query, dict):
        for key, value in query.items():
            term.add_argument(Query('eq', [key, value]))

    if query[0] == '?':
        # raise URIError('Query mustnot start with ?')
        query = query[1:]

    if '/' in query:
        query = re.sub(
            r'[\+\*\$\-:\w%\._]*\/[\+\*\$\-:\w%\._\/]*',
            lambda s: '({})'.format(s.group().replace('/', ',')),
            query
        )

    if JSON_QUERY_COMPATIBLE:
        query = reduce(
            lambda q, args: re.sub(args[0], args[1], q),
            JSON_QUERY_SUBSTITUTIONS.items(),
            query
        )

    query = convert_from_fiql(query)

    current_term = term
    for closed_paran, delim, property_or_value, open_paran in re.findall(PARSING_REGEX, query):
        if delim:
            if delim == '&':
                current_term.set_conjuction('and')
            elif delim == '|':
                current_term.set_conjuction('or')

        if open_paran:
            new_term = Query(property_or_value, parent=current_term)
            current_term.add_argument(new_term)
            if current_term.name in BUBBLE_TO_TOP_OPERATORS:
                top_term.cache[term.name] = term.args
            current_term = new_term
        elif closed_paran:
            if current_term.parent is None:
                raise ParserError('Closing paranthesis without an opening paranthesis')

            # TODO: this isn't clear
            conjunction = current_term.name
            current_term = current_term.parent

            if not conjunction:
                unrequired_query_object = current_term.args.pop()
                current_term.add_argument(unrequired_query_object.args)

        elif (property_or_value or ',' in delim):
            val = str_to_val(property_or_value)
            if val != '':
                current_term.add_argument(val)

            # cache operators we want to bubble to top term
            if term.name in BUBBLE_TO_TOP_OPERATORS:
                top_term.cache[term.name] = term.args

            if term.name == 'eq' and term.args[0] == PRIMARY_KEY_NUMBER:
                top_term.cache[PRIMARY_KEY_NUMBER] = term.args[1]

    if current_term.parent is not None:
        raise ParserError('Opening paranthesis without an closing paranthesis')

    left_over_characters = re.sub(PARSING_REGEX, '', query)
    if left_over_characters:
        # raise URIError('Query must not start with ?')
        pass

    if not top_term.name:
        top_term.name = DEFAULT_NODE_NAME

    return top_term.dict()



#!/usr/bin/env python

# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import socket
from zeroconf_monkey import Zeroconf
from ..logger import Logger
from .mdnsExceptions import InterfaceNotFoundException


class MDNSInterface(object):

    def __init__(self, interfaceIp):
        self.logger = Logger('MDNSInterface')
        self.ip = interfaceIp
        self._establish_interface()

    def _establish_interface(self):
        try:
            self.zeroconf = Zeroconf([self.ip])
        except socket.error:
            msg = "Could not find interface with IP {}".format(self.ip)
            self.logger.writeError(msg)
            raise InterfaceNotFoundException(msg)

    def registerService(self, info):
        self.zeroconf.register_service(info)

    def unregisterService(self, info):
        self.zeroconf.unregister_service(info)

    def close(self):
        self.zeroconf.close()



# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
import re
import logging

FORMAT_SEPERATOR_DEFAULT_REGEX = r'%.'
ISO_FORMAT = '%Y-%m-%dT%H:%M:%S.%fZ'


def split_in_two(x, sep):
    max_split = 1
    try:
        before, after = x.split(sep, max_split)
    except ValueError:
        before, after = x, ''
    return before, after


def partial_match(x, seperators):
    if not seperators:
        return '', []

    sep, remaining_seps = seperators[0], seperators[1:]
    match, rest = split_in_two(x, sep)

    if rest:
        next_match, next_found_seperator = partial_match(rest, remaining_seps)
        if next_match:
            match += sep + next_match
        found_seperators = [sep] + next_found_seperator
    else:
        sep_found = len(match) < len(x)
        found_seperators = [sep] if sep_found else []

    return match, found_seperators


def _partial_match_format(x, format):
    seperators = re.split(FORMAT_SEPERATOR_DEFAULT_REGEX, format)
    seperators = [s for s in seperators if s != '']

    _, found_seperators = partial_match(x, seperators)

    if found_seperators == seperators:
        matched_format = format
    else:
        matched_format, _ = partial_match(format, found_seperators + [seperators[len(found_seperators)]])

    return matched_format


def strptime_partial(x, format=ISO_FORMAT):
    logging.info('looking for partial match with initial format: {}'.format(format))

    matched_format = _partial_match_format(x, format)
    logging.info('found partial format: {} for {}'.format(matched_format, x))

    ret = datetime.strptime(x, matched_format)
    logging.info('returning datetime {}'.format(ret))

    return ret



#!/usr/bin/env python

# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class MDNSAdvertisementCallbackHandler(object):

    def __init__(self, callback, regtype, name, port, txtRecord):
        self.callback = callback
        self.name = name
        self.regtype = regtype
        self.port = port
        self.txtRecord = txtRecord

    def entryCollision(self):
        if self.callback is not None:
            self.callback(self._buildMessage("collision"))

    def entryEstablished(self):
        if self.callback is not None:
            self.callback(self._buildMessage("established"))

    def entryFailed(self):
        if self.callback is not None:
            self.callback(self._buildMessage("failed"))

    def _buildMessage(self, action):
        return {
                "action": action,
                "name": self.name,
                "regtype": self.regtype,
                "port": self.port,
                "txtRecord": self.txtRecord
                }



# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from six import itervalues
from six.moves.urllib.parse import urljoin

import requests
import json
import time

import gevent
import gevent.queue

from .logger import Logger
from .mdnsbridge import IppmDNSBridge
from .mdnsExceptions import ServiceNotFoundException

from .nmoscommonconfig import config as _config
import traceback

AGGREGATOR_APIVERSION = _config.get('nodefacade').get('NODE_REGVERSION')
AGGREGATOR_APINAMESPACE = "x-nmos"
AGGREGATOR_APINAME = "registration"

LEGACY_REG_MDNSTYPE = "nmos-registration"
REGISTRATION_MDNSTYPE = "nmos-register"


class NoAggregator(Exception):
    def __init__(self, mdns_updater=None):
        if mdns_updater is not None:
            mdns_updater.inc_P2P_enable_count()
        super(NoAggregator, self).__init__("No Registration API found")


class InvalidRequest(Exception):
    def __init__(self, status_code=400, mdns_updater=None):
        if mdns_updater is not None:
            mdns_updater.inc_P2P_enable_count()
        super(InvalidRequest, self).__init__("Invalid Request, code {}".format(status_code))
        self.status_code = status_code


class TooManyRetries(Exception):
    def __init__(self, mdns_updater=None):
        if mdns_updater is not None:
            mdns_updater.inc_P2P_enable_count()
        super(TooManyRetries, self).__init__("Too many retries.")


class Aggregator(object):
    """This class serves as a proxy for the distant aggregation service running elsewhere on the network.
    It will search out aggregators and locate them, falling back to other ones if the one it is connected to
    disappears, and resending data as needed."""
    def __init__(self, logger=None, mdns_updater=None):
        self.logger = Logger("aggregator_proxy", logger)
        self.logger.writeWarning("This class is deprecated. Please use the matching one in nmos-node instead.")
        self.mdnsbridge = IppmDNSBridge(logger=self.logger)
        self.aggregator = ""
        self.registration_order = ["device", "source", "flow", "sender", "receiver"]
        self._mdns_updater = mdns_updater
        # 'registered' is a local mirror of aggregated items. There are helper methods
        # for manipulating this below.
        self._registered = {
            'node': None,
            'registered': False,
            'entities': {
                'resource': {
                }
            }
        }
        self._running = True
        self._reg_queue = gevent.queue.Queue()
        self.heartbeat_thread = gevent.spawn(self._heartbeat)
        self.queue_thread = gevent.spawn(self._process_queue)

    # The heartbeat thread runs in the background every five seconds.
    # If when it runs the Node is believed to be registered it will perform a heartbeat
    def _heartbeat(self):
        self.logger.writeDebug("Starting heartbeat thread")
        while self._running:
            heartbeat_wait = 5
            if not self._registered["registered"]:
                self._process_reregister()
            elif self._registered["node"]:
                # Do heartbeat
                try:
                    self.logger.writeDebug("Sending heartbeat for Node {}"
                                           .format(self._registered["node"]["data"]["id"]))
                    self._SEND("POST", "/health/nodes/" + self._registered["node"]["data"]["id"])
                except InvalidRequest as e:
                    if e.status_code == 404:
                        # Re-register
                        self.logger.writeWarning("404 error on heartbeat. Marking Node for re-registration")
                        self._registered["registered"] = False

                        if(self._mdns_updater is not None):
                            self._mdns_updater.inc_P2P_enable_count()
                    else:
                        # Client side error. Report this upwards via exception, but don't resend
                        self.logger.writeError("Unrecoverable error code {} received from Registration API on heartbeat"
                                               .format(e.status_code))
                        self._running = False
                except Exception:
                    # Re-register
                    self.logger.writeWarning("Unexpected error on heartbeat. Marking Node for re-registration")
                    self._registered["registered"] = False
            else:
                self._registered["registered"] = False
                if(self._mdns_updater is not None):
                    self._mdns_updater.inc_P2P_enable_count()
            while heartbeat_wait > 0 and self._running:
                gevent.sleep(1)
                heartbeat_wait -= 1
        self.logger.writeDebug("Stopping heartbeat thread")

    # Provided the Node is believed to be correctly registered, hand off a single request to the SEND method
    # On client error, clear the resource from the local mirror
    # On other error, mark Node as unregistered and trigger re-registration
    def _process_queue(self):
        self.logger.writeDebug("Starting HTTP queue processing thread")
        # Checks queue not empty before quitting to make sure unregister node gets done
        while self._running or (self._registered["registered"] and not self._reg_queue.empty()):
            if not self._registered["registered"] or self._reg_queue.empty():
                gevent.sleep(1)
            else:
                try:
                    queue_item = self._reg_queue.get()
                    namespace = queue_item["namespace"]
                    res_type = queue_item["res_type"]
                    res_key = queue_item["key"]
                    if queue_item["method"] == "POST":
                        if res_type == "node":
                            data = self._registered["node"]
                            try:
                                self.logger.writeInfo("Attempting registration for Node {}"
                                                      .format(self._registered["node"]["data"]["id"]))
                                self._SEND("POST", "/{}".format(namespace), data)
                                self._SEND("POST", "/health/nodes/" + self._registered["node"]["data"]["id"])
                                self._registered["registered"] = True
                                if self._mdns_updater is not None:
                                    self._mdns_updater.P2P_disable()

                            except Exception:
                                self.logger.writeWarning("Error registering Node: %r" % (traceback.format_exc(),))

                        elif res_key in self._registered["entities"][namespace][res_type]:
                            data = self._registered["entities"][namespace][res_type][res_key]
                            try:
                                self._SEND("POST", "/{}".format(namespace), data)
                            except InvalidRequest as e:
                                self.logger.writeWarning("Error registering {} {}: {}".format(res_type, res_key, e))
                                self.logger.writeWarning("Request data: {}".format(data))
                                del self._registered["entities"][namespace][res_type][res_key]

                    elif queue_item["method"] == "DELETE":
                        translated_type = res_type + 's'
                        try:
                            self._SEND("DELETE", "/{}/{}/{}".format(namespace, translated_type, res_key))
                        except InvalidRequest as e:
                            self.logger.writeWarning("Error deleting resource {} {}: {}"
                                                     .format(translated_type, res_key, e))
                    else:
                        self.logger.writeWarning("Method {} not supported for Registration API interactions"
                                                 .format(queue_item["method"]))
                except Exception:
                    self._registered["registered"] = False
                    if(self._mdns_updater is not None):
                        self._mdns_updater.P2P_disable()
        self.logger.writeDebug("Stopping HTTP queue processing thread")

    # Queue a request to be processed. Handles all requests except initial Node POST which is done in _process_reregister
    def _queue_request(self, method, namespace, res_type, key):
        self._reg_queue.put({"method": method, "namespace": namespace, "res_type": res_type, "key": key})

    # Register 'resource' type data including the Node
    # NB: Node registration is managed by heartbeat thread so may take up to 5 seconds!
    def register(self, res_type, key, **kwargs):
        self.register_into("resource", res_type, key, **kwargs)

    # Unregister 'resource' type data including the Node
    def unregister(self, res_type, key):
        self.unregister_from("resource", res_type, key)

    # General register method for 'resource' types
    def register_into(self, namespace, res_type, key, **kwargs):
        data = kwargs
        send_obj = {"type": res_type, "data": data}
        if 'id' not in send_obj["data"]:
            self.logger.writeWarning("No 'id' present in data, using key='{}': {}".format(key, data))
            send_obj["data"]["id"] = key

        if namespace == "resource" and res_type == "node":
            # Handle special Node type
            self._registered["node"] = send_obj
        else:
            self._add_mirror_keys(namespace, res_type)
            self._registered["entities"][namespace][res_type][key] = send_obj
        self._queue_request("POST", namespace, res_type, key)

    # General unregister method for 'resource' types
    def unregister_from(self, namespace, res_type, key):
        if namespace == "resource" and res_type == "node":
            # Handle special Node type
            self._registered["node"] = None
        elif res_type in self._registered["entities"][namespace]:
            self._add_mirror_keys(namespace, res_type)
            if key in self._registered["entities"][namespace][res_type]:
                del self._registered["entities"][namespace][res_type][key]
        self._queue_request("DELETE", namespace, res_type, key)

    # Deal with missing keys in local mirror
    def _add_mirror_keys(self, namespace, res_type):
        if namespace not in self._registered["entities"]:
            self._registered["entities"][namespace] = {}
        if res_type not in self._registered["entities"][namespace]:
            self._registered["entities"][namespace][res_type] = {}

    # Re-register just the Node, and queue requests in order for other resources
    def _process_reregister(self):
        if self._registered.get("node", None) is None:
            self.logger.writeDebug("No node registered, re-register returning")
            return

        try:
            self.logger.writeDebug("Clearing old Node from API prior to re-registration")
            self._SEND("DELETE", "/resource/nodes/" + self._registered["node"]["data"]["id"])
        except InvalidRequest as e:
            # 404 etc is ok
            self.logger.writeInfo("Invalid request when deleting Node prior to registration: {}".format(e))
        except Exception as ex:
            # Server error is bad, no point continuing
            self.logger.writeError("Aborting Node re-register! {}".format(ex))
            return

        self._registered["registered"] = False
        if(self._mdns_updater is not None):
            self._mdns_updater.inc_P2P_enable_count()

        # Drain the queue
        while not self._reg_queue.empty():
            try:
                self._reg_queue.get(block=False)
            except gevent.queue.Queue.Empty:
                break

        try:
            # Register the node, and immediately heartbeat if successful to avoid race with garbage collect.
            self.logger.writeInfo("Attempting re-registration for Node {}"
                                  .format(self._registered["node"]["data"]["id"]))
            self._SEND("POST", "/resource", self._registered["node"])
            self._SEND("POST", "/health/nodes/" + self._registered["node"]["data"]["id"])
            self._registered["registered"] = True
            if self._mdns_updater is not None:
                self._mdns_updater.P2P_disable()
        except Exception as e:
            self.logger.writeWarning("Error re-registering Node: {}".format(e))
            self.aggregator = ""  # Fallback to prevent us getting stuck if the Reg API issues a 4XX error incorrectly
            return

        # Re-register items that must be ordered
        # Re-register things we have in the local cache.
        # "namespace" is e.g. "resource"
        # "entities" are the things associated under that namespace.
        for res_type in self.registration_order:
            for namespace, entities in self._registered["entities"].items():
                if res_type in entities:
                    self.logger.writeInfo("Ordered re-registration for type: '{}' in namespace '{}'"
                                          .format(res_type, namespace))
                    for key in entities[res_type]:
                        self._queue_request("POST", namespace, res_type, key)

        # Re-register everything else
        # Re-register things we have in the local cache.
        # "namespace" is e.g. "resource"
        # "entities" are the things associated under that namespace.
        for namespace, entities in self._registered["entities"].items():
            for res_type in entities:
                if res_type not in self.registration_order:
                    self.logger.writeInfo("Unordered re-registration for type: '{}' in namespace '{}'"
                                          .format(res_type, namespace))
                    for key in entities[res_type]:
                        self._queue_request("POST", namespace, res_type, key)

    # Stop the Aggregator object running
    def stop(self):
        self.logger.writeDebug("Stopping aggregator proxy")
        self._running = False
        self.heartbeat_thread.join()
        self.queue_thread.join()

    def status(self):
        return {"api_href": self.aggregator,
                "api_version": AGGREGATOR_APIVERSION,
                "registered": self._registered["registered"]}

    def _get_api_href(self):
        protocol = "http"
        if _config.get('https_mode') == "enabled":
            protocol = "https"
        api_href = self.mdnsbridge.getHref(REGISTRATION_MDNSTYPE, None, AGGREGATOR_APIVERSION, protocol)
        if api_href == "":
            api_href = self.mdnsbridge.getHref(LEGACY_REG_MDNSTYPE, None, AGGREGATOR_APIVERSION, protocol)
        return api_href

    # Handle sending all requests to the Registration API, and searching for a new 'aggregator' if one fails
    def _SEND(self, method, url, data=None):
        if self.aggregator == "":
            self.aggregator = self._get_api_href()

        headers = None
        if data is not None:
            data = json.dumps(data)
            headers = {"Content-Type": "application/json"}

        url = AGGREGATOR_APINAMESPACE + "/" + AGGREGATOR_APINAME + "/" + AGGREGATOR_APIVERSION + url
        for i in range(0, 3):
            if self.aggregator == "":
                self.logger.writeWarning("No aggregator available on the network or mdnsbridge unavailable")
                raise NoAggregator(self._mdns_updater)

            self.logger.writeDebug("{} {}".format(method, urljoin(self.aggregator, url)))

            # We give a long(ish) timeout below, as the async request may succeed after the timeout period
            # has expired, causing the node to be registered twice (potentially at different aggregators).
            # Whilst this isn't a problem in practice, it may cause excessive churn in websocket traffic
            # to web clients - so, sacrifice a little timeliness for things working as designed the
            # majority of the time...
            try:
                if _config.get('prefer_ipv6') is False:
                    R = requests.request(method, urljoin(self.aggregator, url), data=data, timeout=1.0, headers=headers)
                else:
                    R = requests.request(method, urljoin(self.aggregator, url), data=data, timeout=1.0,
                                         headers=headers, proxies={'http': ''})
                if R is None:
                    # Try another aggregator
                    self.logger.writeWarning("No response from aggregator {}".format(self.aggregator))

                elif R.status_code in [200, 201]:
                    if R.headers.get("content-type", "text/plain").startswith("application/json"):
                        return R.json()
                    else:
                        return R.content

                elif R.status_code == 204:
                    return

                elif (R.status_code/100) == 4:
                    self.logger.writeWarning("{} response from aggregator: {} {}"
                                             .format(R.status_code, method, urljoin(self.aggregator, url)))
                    raise InvalidRequest(R.status_code, self._mdns_updater)

                else:
                    self.logger.writeWarning("Unexpected status from aggregator {}: {}, {}"
                                             .format(self.aggregator, R.status_code, R.content))

            except requests.exceptions.RequestException as ex:
                # Log a warning, then let another aggregator be chosen
                self.logger.writeWarning("{} from aggregator {}".format(ex, self.aggregator))

            # This aggregator is non-functional
            self.aggregator = self._get_api_href()
            self.logger.writeInfo("Updated aggregator to {} (try {})".format(self.aggregator, i))

        raise TooManyRetries(self._mdns_updater)


class MDNSUpdater(object):
    def __init__(self, mdns_engine, mdns_type, mdns_name, mappings, port, logger, p2p_enable=False, p2p_cut_in_count=5,
                 txt_recs=None):
        self.mdns = mdns_engine
        self.mdns_type = mdns_type
        self.mdns_name = mdns_name
        self.mappings = mappings
        self.port = port
        self.service_versions = {}
        self.txt_rec_base = {}
        if txt_recs:
            self.txt_rec_base = txt_recs
        self.logger = logger
        self.logger.writeWarning("This class is deprecated. Please use the matching one in nmos-node instead.")
        self.p2p_enable = p2p_enable
        self.p2p_enable_count = 0
        self.p2p_cut_in_count = p2p_cut_in_count

        for mapValue in itervalues(self.mappings):
            self.service_versions[mapValue] = 0

        self.mdns.register(self.mdns_name, self.mdns_type, self.port, self.txt_rec_base)

        self._running = True
        self._mdns_update_queue = gevent.queue.Queue()
        self.mdns_thread = gevent.spawn(self._modify_mdns)

    def _modify_mdns(self):
        while self._running:
            if self._mdns_update_queue.empty():
                gevent.sleep(0.2)
            else:
                try:
                    txt_recs = self._mdns_update_queue.get()
                    self.mdns.update(self.mdns_name, self.mdns_type, txt_recs)
                except ServiceNotFoundException:
                    self.logger.writeError("Unable to update mDNS record of type {} and name {}"
                                           .format(self.mdns_name, self.mdns_type))

    def stop(self):
        self._running = False
        self.mdns_thread.join()

    def _p2p_txt_recs(self):
        txt_recs = self.txt_rec_base.copy()
        txt_recs.update(self.service_versions)
        return txt_recs

    def update_mdns(self, type, action):
        if self.p2p_enable:
            if (action == "register") or (action == "update") or (action == "unregister"):
                self.logger.writeDebug("mDNS action: {} {}".format(action, type))
                self._increment_service_version(type)
                self._mdns_update_queue.put(self._p2p_txt_recs())

    def _increment_service_version(self, type):
        self.service_versions[self.mappings[type]] = self.service_versions[self.mappings[type]]+1
        if self.service_versions[self.mappings[type]] > 255:
            self.service_versions[self.mappings[type]] = 0

    # Counts up a number of times, and then enables P2P
    def inc_P2P_enable_count(self):
        if not self.p2p_enable:
            self.p2p_enable_count += 1
            if self.p2p_enable_count >= self.p2p_cut_in_count:
                self.P2P_enable()

    def _reset_P2P_enable_count(self):
        self.p2p_enable_count = 0

    def P2P_enable(self):
        if not self.p2p_enable:
            self.logger.writeInfo("Enabling P2P Discovery")
            self.p2p_enable = True
            self._mdns_update_queue.put(self._p2p_txt_recs())

    def P2P_disable(self):
        if self.p2p_enable:
            self.logger.writeInfo("Disabling P2P Discovery")
            self.p2p_enable = False
            self._reset_P2P_enable_count()
            self._mdns_update_queue.put(self.txt_rec_base)
        else:
            self._reset_P2P_enable_count()


if __name__ == "__main__":  # pragma: no cover
    from uuid import uuid4

    agg = Aggregator()
    ID = str(uuid4())

    agg.register("node", ID, id=ID, label="A Test Service", href="http://127.0.0.1:12345/", services=[], caps={},
                 version="0:0", hostname="apiTest")
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        agg.unregister("node", ID)
        agg.stop()



#!/usr/bin/env python

# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .mdnsInterface import MDNSInterface
from .mdnsExceptions import NoNetworkInterfacesFoundException
from ..interfaceController import InterfaceController


class MDNSInterfaceController(object):

    def __init__(self, logger):
        self.logger = logger
        self.interfaces = {}
        self.defaultInterfaces = []
        self._set_default_interfaces()

    def addInterface(self, ip):
        if ip not in self.interfaces.keys():
            self.interfaces[ip] = MDNSInterface(ip)
        return self.interfaces[ip]

    def getInterfaces(self, interfaceIps):
        if interfaceIps == []:
            return self.defaultInterfaces
        else:
            interfaceList = []
            for ip in interfaceIps:
                interfaceList.append(self.addInterface(ip))
                self.logger.writeDebug("Switching to using mDNS interface provided:{}".format(ip))
            return interfaceList

    def close(self):
        for key, interface in self.interfaces.items():
            interface.close()

    def _set_default_interfaces(self):
        ifaceController = InterfaceController(self.logger)
        interfaces = ifaceController.get_default_interfaces()

        if interfaces == []:
            msg = "Could not find any interfaces for mDNS use"
            self.logger.writeError(msg)
            raise NoNetworkInterfacesFoundException(msg)
        for interface in interfaces:
            self.defaultInterfaces.append(self.addInterface(interface))



# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import print_function
import os
import json
import copy

# If updating the defaults, please change them in setup.py too!
config_defaults = {
    "priority": 0,
    "https_mode": "disabled",
    "prefer_ipv6": False,
    "prefer_hostnames": False,
    "node_hostname": None,
    "fix_proxy": "disabled",
    "logging": {
        "level": "DEBUG",
        "fileLocation": "/var/log/nmos.log",
        "output": [
            "file",
            "stdout"
        ]
    },
    "nodefacade": {
        "NODE_REGVERSION": "v1.2"
    },
    "oauth_mode": False,
    "dns_discover": True,
    "mdns_discover": True,
    "interfaces": []
}

config = {}
config.update(copy.deepcopy(config_defaults))

try:
    config_file = "/etc/nmoscommon/config.json"
    if os.path.isfile(config_file):
        f = open(config_file, 'r')
        config_load = json.loads(f.read())
        # Note: This will override objects within the config object in full!
        config.update(config_load)
except Exception as e:
    print("Exception loading config: {}".format(e))



# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from gevent import monkey
monkey.patch_all()

from six.moves import range as xrange # noqa E402
import gevent # noqa E402
import json # noqa E402
import requests # noqa E402
import websocket # noqa E402
import itertools # noqa E402

from .logger import Logger # noqa E402
from .nmoscommonconfig import config as _config # noqa E402

QUERY_APIVERSION = _config.get('nodefacade').get('NODE_REGVERSION')
QUERY_APINAMESPACE = "x-nmos"
QUERY_APINAME = "query"
QUERY_MDNSTYPE = "nmos-query"


class BadSubscriptionError(Exception):
    pass


class QueryNotFoundError(Exception):
    pass


class QueryService(object):

    def __init__(self, mdns_bridge, logger=None, apiversion=QUERY_APIVERSION,  priority=None):
        self.mdns_bridge = mdns_bridge
        self._query_url = self.mdns_bridge.getHref(QUERY_MDNSTYPE, priority)
        iter = 0
        # TODO FIXME: Remove once IPv6 work complete and Python can use link local v6 correctly
        while "fe80:" in self._query_url:
            self._query_url = self.mdns_bridge.getHref(QUERY_MDNSTYPE, priority)
            iter += 1
            if iter > 20:
                break
        self.logger = Logger("nmoscommon.query", logger)
        self.apiversion = apiversion
        self.priority = priority

    def _get_query(self, url):
        backoff = [0.3, 0.7, 1.0]
        for try_i in xrange(len(backoff)):
            try:
                response = requests.get(
                    "{}/{}/{}/{}{}".format(self._query_url, QUERY_APINAMESPACE, QUERY_APINAME, self.apiversion, url)
                )
                response.raise_for_status()
                return response
            except Exception as e:
                self.logger.writeWarning("Could not GET from query service at {}{}: {}".format(self._query_url, url, e))
                if try_i == len(backoff) - 1:
                    raise QueryNotFoundError(e)

                # TODO: sleep between requests to back off
                self._query_url = self.mdns_bridge.getHref(QUERY_MDNSTYPE, self.priority)
                self.logger.writeInfo("Trying query at: {}".format(self._query_url))

        # Shouldn't get this far, but don't return None
        raise QueryNotFoundError("Could not find a query service (should be unreachable!)")  # pragma: no cover

    def get_services(self, service_urn, node_id=None):
        """
        Look for nodes which contain a particular service type.
        Returns a list of found service objects, or an empty list on not-found.
        May raise a QueryNotFound exception if query service can't be contacted.
        """
        response = self._get_query("/nodes/")
        if response.status_code != 200:
            self.logger.writeError("Could not get /nodes/ from query service at {}".format(self._query_url))
            return []

        nodes = response.json()

        services = []

        if node_id is None:
            services = itertools.chain.from_iterable([n.get('services', []) for n in nodes])
        else:
            services = itertools.chain.from_iterable([n.get('services', []) for n in nodes if n["id"] == node_id])

        return [s for s in services if s.get('type', 'unknown') == service_urn]

    def subscribe_topic(self, topic, on_event, on_open=None):
        """
        Subscribe to a query service topic, calling `on_event` for changes.
        Will block unless wrapped in a gevent greenlet:
            gevent.spawn(qs.subscribe_topic, "flows", on_event)
        If `on_open` is given, it will be called when the websocket is opened.
        """
        query_url = self.mdns_bridge.getHref(QUERY_MDNSTYPE, self.priority)

        if query_url == "":
            raise BadSubscriptionError("Could not get query service from mDNS bridge")

        query_url = query_url + "/" + QUERY_APINAMESPACE + "/" + QUERY_APINAME + "/" + self.apiversion

        resource_path = "/" + topic.strip("/")
        params = {"max_update_rate_ms": 100, "persist": False, "resource_path": resource_path, "params": {}}
        r = requests.post(query_url + "/subscriptions", data=json.dumps(params), proxies={'http': ''})
        if r.status_code not in [200, 201]:
            raise BadSubscriptionError("{}: {}".format(r.status_code, r.text))

        r_json = r.json()
        if "ws_href" not in r_json:
            raise BadSubscriptionError("Result has no 'ws_href': {}".format(r_json))

        assert(query_url.startswith("http://"))
        ws_href = r_json.get("ws_href")

        # handlers for websocket events
        def _on_open(*args):
            if on_open is not None:
                on_open()

        def _on_close(*args):
            pass

        def _on_message(*args):
            assert(len(args) >= 1)
            data = json.loads(args[1])
            events = data["grain"]["data"]
            if isinstance(events, dict):
                events = [events]
            for event in events:
                on_event(event)

        # Open websocket connection, and poll
        sock = websocket.WebSocketApp(ws_href, on_open=_on_open, on_message=_on_message, on_close=_on_close)
        if sock is None:
            raise BadSubscriptionError("Could not open websocket at {}".format(ws_href))

        sock.run_forever()


if __name__ == '__main__':  # pragma: no cover
    from .mdnsbridge import IppmDNSBridge
    from pprint import pprint

    qs = QueryService(IppmDNSBridge())

    print(qs.get_services("urn:x-ipstudio:service:status/v1.0"))
    print(qs.get_services("urn:x-ipstudio:service:storedflowquery/v2.0"))
    print(qs.get_services("urn:x-ipstudio:service:fake/v1.0"))

    def callback(data):
        print("\n----\n",)
        pprint(data)

    def on_open():
        print("ONOPEN")

    gevent.spawn(qs.subscribe_topic, "nodes", callback, on_open)

    while True:
        gevent.sleep(1)



# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import absolute_import

import requests
import random

from .nmoscommonconfig import config as _config
from .logger import Logger


class IppmDNSBridge(object):
    def __init__(self, logger=None):
        self.logger = Logger("mdnsbridge", logger)
        self.logger.writeWarning("This class is deprecated. Please use the matching one in nmos-mdns-bridge instead.")
        self.services = {}
        self.config = {}
        self.config.update(_config)

    def _checkLocalQueryServiceExists(self):
        url = "http://127.0.0.1/x-nmos/query/v1.0/"
        try:
            # Request to localhost:18870/ - if it succeeds, the service exists AND is running AND is accessible
            r = requests.get(url, timeout=0.5)
            if r is not None and r.status_code == 200:
                # If any results, put them in self.services
                return url

        except Exception as e:
            self.logger.writeWarning("No local query service running {}".format(e))
        return ""

    def getHref(self, srv_type, priority=None, api_ver=None, api_proto=None):
        if priority is None:
            priority = self.config["priority"]

        if self.logger is not None:
            self.logger.writeDebug("IppmDNSBridge priority = {}".format(priority))

        # Check if type is in services. If not add it
        if srv_type not in self.services:
            self.services[srv_type] = []

        # Check if there are any of that type of service, if not do a request
        no_results = True
        for service in self.services[srv_type]:
            if api_ver is not None and api_ver not in service["versions"]:
                continue
            if api_proto is not None and api_proto != service["protocol"]:
                continue
            if priority >= 100:
                if service["priority"] == priority:
                    no_results = False
            elif service["priority"] < 100:
                no_results = False
        if no_results:
            self._updateServices(srv_type)

        # Re-check if there are any and return "" if not.
        current_priority = 99
        valid_services = []
        for service in self.services[srv_type]:
            if api_ver is not None and api_ver not in service["versions"]:
                continue
            if api_proto is not None and api_proto != service["protocol"]:
                continue
            if priority >= 100:
                if service["priority"] == priority:
                    return self._createHref(service)
            else:
                if service["priority"] < current_priority:
                    current_priority = service["priority"]
                    valid_services = []
                if service["priority"] == current_priority:
                    valid_services.append(service)
        if len(valid_services) == 0:
            self.logger.writeWarning("No services found: {}".format(srv_type))
            if srv_type == "nmos-query":
                return self._checkLocalQueryServiceExists()
            return ""

        # Randomise selection. Delete entry from the services list and return it
        random.seed()
        index = random.randint(0, len(valid_services)-1)
        service = valid_services[index]
        href = self._createHref(service)
        self.services[srv_type].remove(service)
        return href

    def _createHref(self, service):
        proto = service['protocol']
        if service.get('hostname') is not None and self.config["prefer_hostnames"]:
            address = service['hostname']
        else:
            address = service['address']
            if ":" in address:
                address = "[" + address + "]"
        port = service['port']
        return '{}://{}:{}'.format(proto, address, port)

    def _updateServices(self, srv_type):
        req_url = "http://127.0.0.1/x-ipstudio/mdnsbridge/v1.0/" + srv_type + "/"
        try:
            # Request to localhost/x-ipstudio/mdnsbridge/v1.0/<type>/
            r = requests.get(req_url, timeout=0.5, proxies={'http': ''})
            if r is not None and r.status_code == 200:
                # If any results, put them in self.services
                self.services[srv_type] = []
                for dns_data in r.json()["representation"]:
                    if self.config["https_mode"] == "enabled" and dns_data["protocol"] == "https":
                        self.services[srv_type].append(dns_data)
                    elif self.config["https_mode"] != "enabled" and dns_data["protocol"] == "http":
                        self.services[srv_type].append(dns_data)
                    else:
                        self.logger.writeDebug(("Ignoring service with IP {} as protocol '{}' doesn't match the "
                                                "current mode").format(dns_data["address"], dns_data["protocol"]))
        except Exception as e:
            self.logger.writeWarning("Exception updating services: {}".format(e))


if __name__ == "__main__":  # pragma: no cover
    bridge = IppmDNSBridge()
    print(bridge.getHref("nmos-registration"))



#!/usr/bin/env python
from threading import Timer
from zeroconf_monkey import ServiceBrowser


class MDNSSubscription(object):

    def __init__(self, zeroconf, listener, regtype):
        self.listener = listener
        self.regtype = regtype + ".local."
        self.zeroconf = zeroconf
        self._scheduleCallback()
        self.browser = ServiceBrowser(self.zeroconf, self.regtype, self.listener)

    def _scheduleCallback(self):
        self.timer = Timer(3600, self._garbageCollect)
        self.timer.start()

    def _garbageCollect(self):
        self.listener.garbageCollect(self.zeroconf, self.regtype)
        self._scheduleCallback()

    def close(self):
        if self.timer:
            self.timer.cancel()
        self.browser.cancel()



# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import zmq
import threading
import json


class ZmqServer(threading.Thread):
    """ServerTask"""
    def __init__(self, callback, numWorkers=5, host="*", port=5570):
        threading.Thread.__init__(self)
        self.callback = callback
        self.port = port
        self.host = host
        self.numWorkers = numWorkers
        self.workers = []

    def run(self):
        """Sets up an internal router that routes requests to one
        of many worker threads that handle the calls out to the routes"""
        self.context = zmq.Context()
        self.frontend = self.context.socket(zmq.ROUTER)
        self.frontend.bind('tcp://' + self.host + ":" + str(self.port))

        self.backend = self.context.socket(zmq.DEALER)
        self.backend.bind('inproc://backend')

        """Spin up a bunch of workers to handle reqests"""
        for i in range(self.numWorkers):
            worker = ServerWorker(self.callback, self.context)
            worker.start()
            self.workers.append(worker)

        try:
            """This method blocks here until termination"""
            zmq.proxy(self.frontend, self.backend)
        except zmq.ZMQError:
            """Handle cleanly the exception thrown when the context is
            torn down when we finish"""
            pass

    def stop(self):
        for worker in self.workers:
            worker.stop()
            while not worker.is_finished():
                pass
        self.context.destroy()


class ServerWorker(threading.Thread):
    """Serveral instances of these run in different threads to handle
    zmq requests as they come in. This method only wants to be sent JSON,
    please don't send it anything else..."""
    def __init__(self, callback, context):
        threading.Thread.__init__(self)
        self.poller = zmq.Poller()
        self.context = context
        self.amRunning = True
        self.finished = False
        self.callback = callback

    def run(self):
        """Blocking method that runs until amRunning is set false"""
        # Connect to the socket
        worker = self.context.socket(zmq.DEALER)
        self.poller.register(worker, zmq.POLLIN)
        worker.connect('inproc://backend')
        while self.amRunning:
            # Loop round checking for messages until the end
            socks = dict(self.poller.poll(timeout=100))
            if worker in socks:
                # Received a message, do something with it
                ident, msg = worker.recv_multipart()
                try:
                    msg = json.loads(msg)
                except ValueError:
                    response = [400]
                    worker.send_multipart([ident, json.dumps(response)])
                else:
                    response = self.callback(msg)
                    worker.send_multipart([ident, json.dumps(response)])
        self.finished = True
        return

    def stop(self):
        self.amRunning = False

    def is_finished(self):
        return self.finished



from zeroconf_monkey import Zeroconf
from .mdnsSubscription import MDNSSubscription


class MDNSSubscriptionController(object):

    def __init__(self):
        self.zeroconf = Zeroconf()
        self.subscriptions = []

    def addSubscription(self, listener, regtype):
        subscription = MDNSSubscription(self.zeroconf, listener, regtype)
        self.subscriptions.append(subscription)

    def close(self):
        for sub in self.subscriptions:
            sub.close()
        self.subscriptions = []
        self.zeroconf.close()



#!/usr/bin/env python

# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This class is passed a DNS PTR record for a DNS-SD service, and a DNS Listener. It uses
DNSUtils to fetch the service and txt records for the service and uses that information
to notify the dns listener about the new service. It then checks the service twice per TTL
time to ensure it is still there. If it is not, it notifies the DNS listener, then asks the
dnsServiceController to destroy it using a callback passed to it"""


import socket
from . import dnsUtils
from .mdnsExceptions import DNSRecordNotFound
from threading import Timer


class DNSService(object):

    def __init__(
        self,
        pointerRecord,
        type,
        dnsListener,
        removeCallback,
        logger
    ):
        self.pointerRecord = pointerRecord
        self.dnsListener = dnsListener
        self.removeCallback = removeCallback
        self.type = type
        self.name = self.pointerRecord.to_text()
        self.logger = logger
        self._initialiseFromDNS()
        self.running = False

    def _initialiseFromDNS(self):
        self.serviceRecord = dnsUtils.getSRVRecord(self.pointerRecord.to_text())
        self.txtRecord = dnsUtils.getTXTRecord(self.pointerRecord.to_text())
        self.ttl = self.txtRecord.rrset.ttl
        self.txt = self._makeTXTDict(self.txtRecord[0])
        self._parseSRVRecord()

    def _makeTXTDict(self, record):
        toReturn = {}
        bits = record.to_text().split(" ")
        for bit in bits:
            fragments = bit.split("=")
            toReturn[fragments[0][1:]] = fragments[1][:-1]
        return toReturn

    def _removeTrailingDot(self, hostname):
        return hostname.rstrip(".")

    def _parseSRVRecord(self):
        for value in self.serviceRecord:
            entries = value.to_text().split(" ")
            self.hostname = self._removeTrailingDot(entries[3])
            self.address = socket.gethostbyname(self.hostname)
            self.port = int(entries[2])

    def _removeService(self):
        self.dnsListener.removeListener(self)
        self.ttlTimer.cancel()

    def _updateService(self):
        self.dnsListener.removeListener(self)
        self.dnsListener.addListener(self)

    def _addService(self):
        self.dnsListener.addListener(self)

    def _ttlTimerCallback(self):
        try:
            self._initialiseFromDNS()
            self._updateService()
        except DNSRecordNotFound:
            self.removeCallback(self)
        else:
            self._startTimer()

    def _startTimer(self):
        self.ttlTimer = Timer(self.ttl, self._ttlTimerCallback)
        self.ttlTimer.start()

    def start(self):
        self._startTimer()
        self.running = True
        self._addService()

    def stop(self):
        if self.running:
            self.ttlTimer.cancel()
        self.running = False

    def close(self, signum=None, frame=None):
        self.dnsListener.removeListener(self)
        self.stop()



# Copyright 2019 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import requests
import json

from os import path, sep
from functools import wraps
from time import time
from flask import request
from functools import reduce
from six.moves.urllib.parse import urljoin, parse_qs
from requests.exceptions import RequestException
from authlib.jose import jwt, jwk
from authlib.common.errors import AuthlibBaseError
from authlib.oauth2.rfc6749.errors import MissingAuthorizationError, UnsupportedTokenTypeError
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization

from ..mdnsbridge import IppmDNSBridge
from ..nmoscommonconfig import config as _config
from ..logger import Logger as defaultLogger
from .claims_options import IS_XX_CLAIMS
from .claims_validator import JWTClaimsValidator

MDNS_SERVICE_TYPE = "nmos-auth"
OAUTH_MODE = _config.get('oauth_mode', True)

NMOSAUTH_DIR = path.abspath(path.join(sep, 'var', 'nmosauth'))
CERT_FILE = 'certificate.pem'
CERT_FILE_PATH = path.join(NMOSAUTH_DIR, CERT_FILE)

APINAMESPACE = "x-nmos"
APINAME = "auth"
APIVERSION = "v1.0"
JWK_ENDPOINT = 'jwks'
JWK_URL_PATH = '/{}/{}/{}/{}'.format(APINAMESPACE, APINAME, APIVERSION, JWK_ENDPOINT)

REFRESH_KEY_INTERVAL = 3600


class RequiresAuth(object):

    def __init__(self, condition=OAUTH_MODE, claimsOptions=IS_XX_CLAIMS):
        self.logger = defaultLogger("authresource")
        self.bridge = IppmDNSBridge()
        self.condition = condition
        self.claimsOptions = claimsOptions
        self.publicKey = None
        self.key_last_accessed_time = 0

    def getHrefFromService(self, serviceType):
        return self.bridge.getHref(serviceType)

    def getJwksFromEndpoint(self):
        try:
            href = self.getHrefFromService(MDNS_SERVICE_TYPE)
            jwk_href = urljoin(href, JWK_URL_PATH)
            self.logger.writeInfo('JWK endpoint is: {}'.format(jwk_href))
            jwk_resp = requests.get(jwk_href, timeout=0.5, proxies={'http': ''})
            jwk_resp.raise_for_status()  # Raise error if status !=200
        except RequestException as e:
            self.logger.writeError("Error: {0!s}".format(e))
            self.logger.writeError("Cannot find certificate at {}. Is the Auth Server Running?".format(jwk_href))
            raise

        if "application/json" in jwk_resp.headers['content-type']:
            try:
                jwks = jwk_resp.json()
                if "keys" in jwks:
                    jwks_keys = jwks["keys"]
                    self.logger.writeInfo("JSON Web Key Set located at /jwks endpoint.")
                else:
                    jwks_keys = jwks
                return jwks_keys
            except Exception as e:
                self.logger.writeError("Error: {}. JWK endpoint contains: {}".format(str(e), jwk_resp.text))
                raise
        else:
            self.logger.writeError("Incorrect Content-Type. Expected 'application/json but got {}".format(
                jwk_resp.headers['content-type']))
            raise ValueError

    def getCertFromFile(self, file_path):
        try:
            if path.exists(file_path):
                with open(file_path, 'r') as myfile:
                    cert = myfile.read()
                    return cert
        except OSError:
            self.logger.writeError("File {} does not exist or you do not have permission to open it".format(
                file_path))
            raise

    def getPublicKeyString(self, key_class):
        serialised_key = key_class.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )
        return serialised_key.decode('utf-8')

    def findMostRecentJWK(self, jwks):
        """Finds the JWK with the most recent timestamp inside the 'kid' property"""
        try:
            newest_key = reduce(
                lambda a, b: a if int(a["kid"].lstrip("x-nmos-")) > int(b["kid"].lstrip("x-nmos-")) else b, jwks)
            return newest_key
        except KeyError as e:
            self.logger.writeError("Format of JSON Web Key 'kid' parameter is incorrect. {}".format(e))
            raise

    def extractPublicKey(self, key_containing_object):
        """Extracts a key from the given parameter. A list or a dict object will be treated as a JWK or JWKS structure.
        A string will be treated like a X509 certificate"""
        try:
            if isinstance(key_containing_object, dict):
                key_class = jwk.loads(key_containing_object)
            elif isinstance(key_containing_object, list):
                newest_key = self.findMostRecentJWK(key_containing_object)
                key_class = jwk.loads(newest_key)
            elif isinstance(key_containing_object, str):
                key_containing_object = key_containing_object.encode()
                crt_obj = x509.load_pem_x509_certificate(key_containing_object, default_backend())
                key_class = crt_obj.public_key()
            pubkey_string = self.getPublicKeyString(key_class)
            if pubkey_string is None:
                self.logger.writeError("Public Key could not be extracted from certificate")
            else:
                return pubkey_string
        except Exception as e:
            self.logger.writeError("Public Key(s) could not be extracted from JSON Web Keys. {}".format(e))
            raise

    def getPublicKey(self):
        if self.publicKey is None or self.key_last_accessed_time + REFRESH_KEY_INTERVAL < time():
            try:
                self.logger.writeInfo("Fetching JSON Web Keys using DNS Service Discovery")
                jwks = self.getJwksFromEndpoint()
                public_key = self.extractPublicKey(jwks)
            except Exception as e:
                self.logger.writeError("Error: {0!s}. Trying to fetch certificate from file".format(e))
                cert = self.getCertFromFile(CERT_FILE_PATH)
                public_key = self.extractPublicKey(cert)
            self.publicKey = public_key
            self.key_last_accessed_time = time()
        return self.publicKey

    def processAccessToken(self, auth_string):
        # Auth string is of type 'Bearer xAgy65..'
        if auth_string.find(' ') > -1:
            token_type, token_string = auth_string.split(None, 1)
            if token_type.lower() != "bearer":
                raise UnsupportedTokenTypeError()
        # Otherwise string is access token 'xAgy65..'
        else:
            token_string = auth_string
        if token_string == "null" or token_string == "":
            raise MissingAuthorizationError()
        pubKey = self.getPublicKey()
        claims = jwt.decode(s=token_string, key=pubKey,
                            claims_cls=JWTClaimsValidator,
                            claims_options=self.claimsOptions,
                            claims_params=None)
        claims.validate()

    def handleHttpAuth(self):
        """Handle bearer token string ("Bearer xAgy65...") in "Authorzation" Request Header"""
        auth_string = request.headers.get('Authorization', None)
        if auth_string is None:
            raise MissingAuthorizationError()
        self.processAccessToken(auth_string)

    def handleSocketAuth(self, *args, **kwargs):
        """Handle bearer token string ("access_token=xAgy65...") in Websocket URL Query Param"""
        ws = args[0]
        environment = ws.environ
        auth_header = environment.get('HTTP_AUTHORIZATION', None)
        if auth_header is not None:
            self.logger.writeInfo("auth header string is {}".format(auth_header))
            auth_string = auth_header
            self.processAccessToken(auth_string)
        else:
            self.logger.writeWarning("Websocket does not have auth header, looking in query string..")
            query_string = environment.get('QUERY_STRING', None)
            try:
                if query_string is not None:
                    auth_string = parse_qs(query_string)['access_token'][0]
                    self.processAccessToken(auth_string)
                else:
                    raise MissingAuthorizationError
            except (AuthlibBaseError, KeyError):
                err = {"type": "error", "data": "Socket Authentication Error"}
                ws.send(json.dumps(err))
                self.logger.writeError("""
                    'access_token' URL param doesn't exist. Websocket authentication failed.
                """)
                raise

    def JWTRequired(self):
        def JWTDecorator(func):
            @wraps(func)
            def processAccessToken(*args, **kwargs):
                # Check to see if request is a Websocket upgrade, else treat request as a standard HTTP request
                headers = request.headers
                if ('Upgrade' in headers.keys() and headers['Upgrade'].lower() == 'websocket'):
                    self.logger.writeInfo("Using Socket handler")
                    self.handleSocketAuth(*args, **kwargs)
                else:
                    self.logger.writeInfo("Using HTTP handler")
                    self.handleHttpAuth()
                return func(*args, **kwargs)
            return processAccessToken
        return JWTDecorator

    def __call__(self, func):
        if not self.condition:
            # Return the function unchanged, not decorated.
            return func

        # Return decorated function
        decorator = self.JWTRequired()
        return decorator(func)



#!/usr/bin/env python

# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from socket import inet_ntoa
from ..logger import Logger
from six.moves import queue
from threading import Thread


class MDNSListener(object):

    def __init__(self, callbackMethod, registerOnly):
        self.logger = Logger("zeroconf mDNS")
        self.callback = callbackMethod
        self.registerOnly = registerOnly
        self.records = {}
        self.nameMap = {}
        self._resolve_queue = queue.Queue()

    def add_service(self, zeroconf, srv_type, name):
        self._resolve_queue.put((srv_type, name, zeroconf))
        resolveThread = Thread(target=self._resolve_service_details)
        resolveThread.daemon = True
        resolveThread.start()

    def _resolve_service_details(self):
        threadParameters = self._resolve_queue.get()
        srv_type = threadParameters[0]
        name = threadParameters[1]
        zeroconf = threadParameters[2]
        info = zeroconf.get_service_info(srv_type, name)
        if info is not None:
            self._update_record(srv_type, name, info)
            self._respondToClient(name, "add", info)
        elif name in self.records[srv_type]:
            self.remove_service(zeroconf, srv_type, name)
        self._resolve_queue.task_done()

    def _update_record(self, srv_type, name, info):
        try:
            self.records[srv_type][name] = info
        except KeyError:
            self.records[srv_type] = {}
            self.records[srv_type][name] = info

    def remove_service(self, zeroconf, srv_type, name):
        info = self.records[srv_type].pop(name)
        if not self.registerOnly:
            action = "remove"
            self._respondToClient(name, action, info)

    def _respondToClient(self, name, action, info):
        callbackData = self._buildClientCallback(action, info)
        self.logger.writeDebug("mDNS Service {}: {}".format(action, callbackData))
        self.callback(callbackData)

    def _buildClientCallback(self, action, info):
        return {
            "action": action,
            "type": info.type,
            "name": info.name,
            "port": info.port,
            "hostname": info.server.rstrip("."),
            "address": inet_ntoa(info.address),
            "txt": info.properties
        }

    def garbageCollect(self, zeroconf, srv_type):
        for name in list(self.records[srv_type].keys()):
            self._resolve_queue.put((srv_type, name, zeroconf))
            resolveThread = Thread(target=self._resolve_service_details)
            resolveThread.daemon = True
            resolveThread.start()



# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import logging
import sys
from .nmoscommonconfig import config as _config

# Check to see if BBC R&D internal logger is present
try:  # pragma: no cover
    from pyipputils.ipplogger import IppLogger
    IPP_LOGGER = True
except ImportError:  # pragma: no cover
    IPP_LOGGER = False

__all__ = ["Logger"]


class PurePythonLogger:

    logsOpened = False

    def __init__(self, node_name='nmos-root', _parent=None):
        logFormat = logging.Formatter(
            '%(asctime)s : %(name)s : %(levelname)s : %(message)s')

        logLevel = getattr(
            logging,
            _config.get("logging").get("level"))
        logOutputs = _config.get("logging").get("output")

        self.log = logging.getLogger(node_name)
        self.log.setLevel(logLevel)

        if not PurePythonLogger.logsOpened:
            if "file" in logOutputs:
                try:
                    fileLoc = _config.get(
                        "logging").get(
                        "fileLocation")
                    fileHandler = logging.FileHandler(fileLoc)
                    fileHandler.setLevel(logLevel)
                    fileHandler.setFormatter(logFormat)
                    self.log.addHandler(fileHandler)
                except Exception:
                    pass

            if "stdout" in logOutputs:
                streamHandler = logging.StreamHandler(sys.stdout)
                streamHandler.setLevel(logLevel)
                streamHandler.setFormatter(logFormat)
                self.log.addHandler(streamHandler)

            if "stderr" in logOutputs:
                streamHandler = logging.StreamHandler(sys.stderr)
                streamHandler.setLevel(logLevel)
                streamHandler.setFormatter(logFormat)
                self.log.addHandler(streamHandler)

            self.log.propagate = False

            PurePythonLogger.logsOpened = True

        logging.info("Logging started...")

    def writeFatal(self, message):
        self.log.fatal(message)

    def writeError(self, message):
        self.log.error(message)

    def writeWarning(self, message):
        self.log.warning(message)

    def writeInfo(self, message):
        self.log.info(message)

    def writeDebug(self, message):
        self.log.debug(message)


# Use BBC R&D logger in preference to this one
if True:  # pragma: no cover
    if IPP_LOGGER:
        Logger = IppLogger
    else:
        Logger = PurePythonLogger



#!/usr/bin/env python

# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .dnsListener import DNSListener
from .dnsService import DNSService
from . import dnsUtils
from .mdnsExceptions import DNSRecordNotFound, NoDefaultDnsSearchDomain
from threading import Timer


"""This class uses the DNSUtils class to check that the current domain
has the PTR record for DNS-SD, and then finds all the available services
and creates instances of dnsService to watch them. It then checks back to
see if there are new services at least twice in the duration of the shortest
TTL of any found service, or no less than once an hour, whichever is shorter"""


class DNSServiceController(object):

    def __init__(self, type, callback, logger, registerOnly, domain=None):
        self.running = False
        self.logger = logger
        self.type = type
        self.services = {}
        self.domain = domain
        self.listener = DNSListener(callback, registerOnly)
        self.timer = None

    def start(self):
        self._scheduleCallback()
        self.running = True

    def _scheduleCallback(self):
        if self.running:
            interval = self._findServiceRefreshInterval()
        else:
            interval = 1
        self.timer = Timer(interval, self._checkForServiceUpdatesCallback)
        self.timer.start()

    def _findServiceRefreshInterval(self):
        accumulator = 3600  # 3600 seconds = 1 hour
        for service in self.services.values():
            target = service.ttl/2
            if target < accumulator:
                accumulator = target
        return accumulator

    def _getDNSServices(self):
        try:
            if not dnsUtils.checkDNSSDActive(self.domain):
                self.logger.writeError("DNS-SD pointer record not found on current domain")
                return []
        except NoDefaultDnsSearchDomain:
            self.logger.writeError("No default DNS search domain found")
            return []

        try:
            return dnsUtils.discoverService(self.type, self.domain)
        except DNSRecordNotFound:
            self.logger.writeDebug("Could not find any DNS services of type {}".format(self.type))
            return []

    def _populateServices(self, pointerRecords):
        for record in pointerRecords:
            service = DNSService(
                record,
                self.type,
                self.listener,
                self._removeServiceCallback,
                self.logger
            )
            if service.name not in self.services:
                self.services[service.name] = service
                service.start()

    def _removeServiceCallback(self, serviceToRemove):
        serviceToRemove.close()
        self.services.pop(serviceToRemove.name)

    def _checkForServiceUpdatesCallback(self):
        serviceRecords = self._getDNSServices()
        self._populateServices(serviceRecords)
        self._scheduleCallback()

    def close(self):
        for _, service in self.services.items():
            service.close()
        if self.timer:
            self.timer.cancel()



# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import print_function
from six.moves.urllib.parse import unquote
from .dateutil import strptime_partial
from datetime import datetime
import json
import re
import logging

JSON_QUERY_COMPATIBLE = True

CONVERT_TYPES = {
    'true': True,
    'false': False,
    'none': None,
    'null': None,
    'undefined': None,
    'infinity': float('inf'),
    '-infinity': -float('inf')
}


class ConversionError(Exception):
    pass


def converter_auto(string):
    if string.lower() in CONVERT_TYPES:
        known_conversion = CONVERT_TYPES[string.lower()]
        logging.info('auto conversition for {} -> {}'.format(string, known_conversion))
        ret = known_conversion
    elif string.isdigit():
        logging.info('auto converting {} to integer'.format(string))
        ret = int(string)
    else:
        ret = unquote(string)
        if ret and JSON_QUERY_COMPATIBLE:
            if "'" == ret[0] == ret[-1]:
                ret = json.loads('"{}"'.format(ret[1: -1]))
    if ret == string:
        logging.info('auto converting made no change to {}'.format(string))
    return ret


def converter_number(n):
    return int(n)


def converter_epoch(x):
    delta = (strptime_partial(x) - datetime(1970, 1, 1))
    return int(1000 * delta.total_seconds())


def converter_isodate(x):
    return strptime_partial(x)


def converter_date(x):
    return strptime_partial(x)


def converter_boolean(x):
    x = x.lower()
    if x == 'true':
        return True
    elif x == 'false':
        return False
    else:
        raise ConversionError('Do not recognise boolean: {}'.format(x))
    return x == 'true'


def converter_string(x):
    return unquote(x)


def converter_re(x):    # TODO: no test coverage
    return re.compile(converter_string(x), flags=re.IGNORECASE)


def converter_RE(x):    # TODO: no test coverage
    return re.compile(converter_string(x))


def converter_glob(x):
    raise NotImplementedError('cannot convert blob {}'.format(x))


CONVERTERS = {
    'auto': converter_auto,
    'number': converter_number,
    'epoch': converter_epoch,
    'isodate': converter_isodate,
    'date': converter_date,
    'boolean': converter_boolean,
    'string': converter_string,
    're': converter_re,
    'RE': converter_RE,
    'glob': converter_glob
}

FIQL_REGEX = r'(\([\+\*\$\-:\w%\._,]+\)|[\+\*\$\-:\w%\._]*|)([<>!]?=(?:[\w]*=)?|>|<)(\([\+\*\$\-:\w%\._,]+\)|[\+\*\$\-:\w%\._]*|)'

OPERATOR_MAP = {
    '=': 'eq',
    '==': 'eq',
    '>': 'gt',
    '>=': 'ge',
    '<': 'lt',
    '<=': 'le',
    '!=': 'ne'
}


def convert_from_fiql(query):

    def substitutor(match):
        prop, operator, value = match.groups()
        if len(operator) < 3:
            try:
                operator = OPERATOR_MAP[operator]
            except KeyError:
                raise ConversionError("Illegal operator {}".format(operator))
        else:
            operator = operator[1:-1]
        return operator + '(' + prop + ',' + value + ')'

    return re.sub(FIQL_REGEX, substitutor, query)



# Copyright 2017 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import netifaces

from nmoscommon import nmoscommonconfig
try:
    from pyipputils import ipphostname
    ippUtilsAvailable = True
except ImportError:
    ippUtilsAvailable = False
    ipphostname = None  # used to make test patching work


class InterfaceController():

    def __init__(self, logger):
        self.logger = logger

    def get_default_interfaces(self):
        """Returns a list of interface IPs"""
        interfaces = self._get_default_interface_from_file()
        if interfaces == []:
            interfaces = self._get_default_interface_from_routing_rules()
        if interfaces == []:
            interfaces = self._get_default_interface_from_gateway()
        if interfaces == []:
            interfaces = self._get_all_interfaces()
        if interfaces == []:
            msg = "Could not find any interfaces"
            self.logger.writeError(msg)
        return interfaces

    def _get_all_interfaces(self):
        interfaces = []
        for interface in netifaces.interfaces():
            if (interface is not None) & (str(interface)[0:2] != 'lo'):
                try:
                    ipv4Addr = netifaces.ifaddresses(interface)[netifaces.AF_INET][0]['addr']
                    if str(ipv4Addr)[0:4] != "127.":
                        interfaces.append(ipv4Addr)
                except (KeyError, IndexError):
                    pass
        if interfaces != []:
            self.logger.writeDebug("Using all available IPv4 network interfaces:"
                                   "{}".format(interfaces))
        return interfaces

    def _get_default_interface_from_routing_rules(self):
        if ippUtilsAvailable:
            interfaces = [ipphostname.getLocalIP()]
            self.logger.writeDebug("Choosing interface using routing rules:"
                                   "{}".format(interfaces[0]))
            return interfaces
        else:
            self.logger.writeDebug("Could not find ipp-utils, will try using default gateway interface")
            return []

    def _get_default_interface_from_gateway(self):
        defaultGateways = netifaces.gateways()
        try:
            defaultInterfaceName = defaultGateways['default'][netifaces.AF_INET][1]
            defaultInterfaceDetails = netifaces.ifaddresses(defaultInterfaceName)
            defaultInterfaceIp = defaultInterfaceDetails[netifaces.AF_INET][0]['addr']
        except KeyError:
            return []
        self.logger.writeDebug("Choosing interface using the default gateway:"
                               " {}".format(defaultInterfaceIp))
        return [defaultInterfaceIp]

    def _get_default_interface_from_file(self):
        try:
            interfaces = []
            interfaceNames = nmoscommonconfig.config['interfaces']
            if interfaceNames == "*":
                return self._get_all_interfaces()
            for interfaceName in interfaceNames:
                interfaces.append(netifaces.ifaddresses(interfaceName)[netifaces.AF_INET][0]['addr'])
            self.logger.writeDebug("Choosing interface from /etc/nmoscommon/config.json "
                                   "file: {}".format(interfaces))
        except KeyError:
            self.logger.writeDebug("No interface config file - will try and use routing rules")

        return interfaces
